# -----------------------------------------------------------------------------
# Copyright (c) 2020-2024 C. Aubert, T. Rubiano, N. Rusch and T. Seiller.
#
# This file is part of pymwp.
#
# pymwp is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pymwp is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pymwp. If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# flake8: noqa: W605

from __future__ import annotations

import logging
from functools import reduce
from typing import Optional, List, Tuple, Union

from . import DELTAS, Monomial
from .constants import Comparison, SetInclusion
from .semiring import ZERO_MWP, INFTY_MWP, sum_mwp

logger = logging.getLogger(__name__)


class Polynomial:
    """A polynomial is an ordered list of ordered Monomials.

    For polynomials, I introduce a total order on the monomials. This
    eases the computation of the sum: if we want to add a monomial to an
    ordered list of monomials, we compare the monomial to each of the
    elements of the list until we find either an element which is equal
    (and then we sum the scalars) or an element which is larger (and
    then we insert the new monomial there).

    Polynomials use the following ordering: $\\delta(i,j)$
    is smaller than $\\delta(m,n)$ iff either $j<n$ or $(j==n)$ and $(i<m)$.

    This is extended to products (which we consider ordered!) by
    letting $\\prod_k\\delta(i_k,j_k) < \\prod_l\\delta(m_l,n_l)$
    iff $\\delta(i_1,j_1) < \\delta(m_1,n_1)$.

    Attributes:
         list (List[Monomial]): List of monomials.
    """

    def __init__(
            self,
            *monomials: Optional[Union[str, Monomial, Tuple[str, DELTAS]]]
    ):
        """Create a polynomial.

        Example:
            Create polynomial with 0-monomial:

            ```python
            zero = Polynomial()
            ```

            Create polynomial with one monomial with specific scalar:

            ```python
            poly = Polynomial('w')               # shorthand
            poly = Polynomial(Monomial('w'))     # longer, equivalent
            ```

            Create polynomial with two monomials and lists of deltas:

            ```python
            poly = Polynomial(('m', (0, 1)), ('w', (0, 0), (1, 1)))
            ```

        Arguments:
            monomials: arbitrary monomials.
        """
        m_list = [Monomial.format(v) for v in monomials]
        self.list = m_list if len(m_list) > 0 else [Monomial(ZERO_MWP)]

    def __str__(self):
        values = ''.join(['+' + str(m) for m in self.list]) or ('+' + ZERO_MWP)
        return " " + values

    def __eq__(self, other):
        return self.equal(other)

    def __add__(self, other):
        return self.add(other)

    def __mul__(self, other):
        return self.times(other)

    def eval(self, *scalars: str) -> List[DELTAS]:
        """List of monomial deltas with scalar in *scalars.
        Scalars always includes $\\infty$, but can include other flows."""
        match = scalars + (INFTY_MWP,)
        return [tuple(mono.deltas) for mono in
                [m for m in self.list if m.scalar in match]]

    @staticmethod
    def inclusion(list_monom: list, mono: Monomial, i: int = 0) \
            -> Tuple[bool, int]:
        """Filter list_monom regarding mono inclusion and return info.

        Remove all monomials of list_monom that are included in mono.

        Return CONTAINS if one of monomials of list_monom contains mono
        (regarding Monomial.inclusion def).

        Arguments:
            list_monom: A list of monomials.
            mono: A monomial we want to add.
            i: The position index where to add mono.

        Returns:
            False if mono already in list_monom and shifted index where to
                insert mono, return True if mono not in list_monom.
        """
        j = 0
        while j < len(list_monom):
            m = list_monom[j]
            incl = m.inclusion(mono)
            # if m ⊆ mono
            if incl == SetInclusion.CONTAINS:
                # We will then add mono so we can remove m
                list_monom.remove(m)
                # If removed monom is before i (where we want to insert mono)
                if j < i:
                    i = i - 1  # shift left position
                continue
            elif incl == SetInclusion.INCLUDED:
                # We don't want to add mono, inform with CONTAINS
                return False, i
            j = j + 1
        # No inclusion
        return True, i

    def add(self, polynomial: Polynomial) -> Polynomial:
        """Add two polynomials.

        - If both lists are empty the result is empty.
        - If one list is empty, the result will be the other list
        of polynomials.

        Otherwise, the operation will zip the two lists together and
        return a new polynomial of sorted monomials.

        Arguments:
            polynomial: Polynomial to add to self.

        Returns:
            New, sorted polynomial that is a sum of the two input polynomials.
        """
        # check for empty lists
        if not self.list and not polynomial.list:
            return Polynomial()
        if not self.list:
            return polynomial.copy()
        if not polynomial.list:
            return self.copy()

        i, j = 0, 0
        new_list = self.copy().list
        # self_len = len(new_list)
        poly_len = len(polynomial.list)

        # iterate lists of monomials until the end of shorter list
        while j < poly_len:
            mono2 = polynomial.list[j]

            tobe_inserted, i = Polynomial.inclusion(new_list, mono2, i)

            if not tobe_inserted:
                j = j + 1
                continue

            # handle case where first list is shorter
            # by just appending what remains of the
            # other list of monomials
            if i == len(new_list):
                for m in polynomial.list[j:]:
                    tobe_inserted, i = Polynomial.inclusion(new_list, m, i)
                    if tobe_inserted:
                        new_list = new_list + [m]
                break

            mono1 = new_list[i]

            check = Polynomial.compare(mono1.deltas, mono2.deltas)

            # move to next when self is smaller
            if check == Comparison.SMALLER:
                i = i + 1

            # insert when the second is smaller
            elif check == Comparison.LARGER:
                new_list.insert(i, mono2)
                i = i + 1
                j = j + 1

            # when both list heads are the same
            # recompute scalar and move to next element
            else:
                new_list[i].scalar = sum_mwp(mono1.scalar, mono2.scalar)
                j = j + 1

        sorted_monomials = Polynomial.sort_monomials(new_list)
        return Polynomial(*sorted_monomials).remove_zeros()

    def times(self, polynomial: Polynomial) -> Polynomial:
        """Multiply two polynomials.

        Here we assume at least self is a sorted polynomial,
        and the result of this operation will be sorted.

        This operation works as follows:

        1. We compute a table of all the separated products
            $P.m_1,...,P.m_n$. Each of the elements is itself
            a sorted list of monomials: $P.m_j=m^j_1,...,m^j_k$

        2. We then sort the list of the first (smallest) elements
            of each list. I.e. we sort the list $m^1_1,m^2_1,...,m^n_1$
            and produce the list corresponding list of indexes of
            length n, I.e. a permutation over ${0,...,n}$.

        3. Once all this preparatory operations are done, the main part
           of the algorithm goes as follows:

        4. We consider the first element — say j — of the list of indexes
           and append to the result the first element of the corresponding
           list $P.m_j$ of monomials.

        5. We remove both the first element of the list of index and
           the first element of $P.m_j$.

        6. If $P.m_j$ is not empty, we insert j in the list of index
           at the right position: for this we compare the (new) first
           element of $P.m_j$ to  $m^{i_2}_1$ (as we removed the
           first element, $i_2$ is now the head of the list of indexes),
           then $m^{i_3}_1$, until we reach the index h such that
           $m^{i_h}_1$ is larger than $m^{j}_1$.

        7. We start back at point 4. Unless only one element is left
           in the list of indexes. In this case, we simply append the
           tail of the corresponding list to the result.

        Arguments:
            polynomial: polynomial to multiply with self.

        Returns:
            A new polynomial that is the sorted product of the two input
                polynomials.
        """

        # 1: compute table of products
        # here we compute P1 x P2 for each polynomial, excluding from the
        # result all monomials that have scalar value 0
        products = [[mono for mono in (m1 * m2 for m1 in self.list)
                     if mono.scalar != ZERO_MWP] for m2 in polynomial.list]
        # filter out empty monomials
        table: List[List[Monomial]] = [p for p in products if p]

        # if table is empty, return zero polynomial
        if not table:
            return Polynomial()

        # 2: create an index lists that represents the ordered
        # monomials in table, ordered by deltas of first monomials
        index_list = [0]
        for i in range(1, len(table)):
            t1 = table[i][0].deltas
            for j in range(len(index_list)):
                t2 = table[index_list[j]][0].deltas
                if Polynomial.compare(t1, t2) == Comparison.SMALLER:
                    index_list.insert(j, i)
                    break
            if i not in index_list:
                index_list.append(i)

        # 3: start main part
        result = []
        while index_list:
            # 4. get first element and append to result
            # 5. remove from index and table
            smallest = index_list.pop(0)
            mono2 = table[smallest].pop(0)
            tobe_inserted, _ = Polynomial.inclusion(result, mono2)
            if tobe_inserted:
                result.append(mono2)

            # 6. when table is non-empty insert j at
            # the right index
            if table[smallest]:
                inserted = False
                t1 = table[smallest][0].deltas
                for j in range(len(index_list)):
                    t2 = table[index_list[j]][0].deltas
                    if Polynomial.compare(t1, t2) == Comparison.SMALLER:
                        index_list.insert(j, smallest)
                        inserted = True
                        break
                if not inserted:
                    index_list.append(smallest)
            # 7. repeat until done

        return Polynomial(*result).remove_zeros()

    def equal(self, polynomial: Polynomial) -> bool:
        """Determine if two polynomials are equal.

        This method will compare current polynomial (self) to
        another polynomial provided as argument. Result of
        true means both polynomials have an equal number of
        monomials, and element-wise each monomial has the same
        list of deltas. Otherwise, the result is false.

        This method is alias of `==` operator.

        Arguments:
            polynomial: polynomial to compare.

        Returns:
            True if polynomials are equal and false otherwise.
        """
        p1, p2 = self.list, polynomial.list

        # the only times deltas are equal is if they contain
        # same values and are equal in length; avoid calling
        # compare because it is more expensive method call; we
        # can do faster equality comparison on deltas this way
        same = [m1.scalar == m2.scalar and m1.deltas == m2.deltas
                for m1, m2 in zip(p1, p2)]

        # if False is in list it means some comparison of
        # deltas was determined not to be equal; do length
        # comparison last because it is almost never False
        return False not in same and len(p1) == len(p2)

    def copy(self) -> Polynomial:
        """Make a deep copy of polynomial."""
        return Polynomial(*[m.copy() for m in self.list])

    def show(self) -> None:
        """Display polynomial."""
        print(str(self))

    @property
    def some_infty(self) -> bool:
        """True if some monomial yields an infinity choice."""
        for mono in self.list:
            if mono.scalar == 'i':
                return True
        return False

    def choice_scalar(self, *choices: int, least_scalar: str = None) \
            -> Optional[str]:
        """For given sequence of choices, determine corresponding scalar.

        Arguments:
            choices: tuple of choices.
            least_scalar: typically zero, but can be m on the diagonal.

        Returns:
            Scalar value matching choices or None.
        """
        scalars = [mono.choice_scalar(*choices) for mono in self.list]
        scalars = [scalar for scalar in scalars if scalar]  # exclude None
        return reduce(sum_mwp, scalars) if scalars else least_scalar

    @staticmethod
    def compare(delta_list1: list, delta_list2: list) -> Comparison:
        """
        Compare 2 lists of deltas.

        We compare the initial segment up to the size of the shortest one.
        If the initial segments match, then the result is determined based
        on length. Three outputs are possible:

        - `SMALLER` if the first list is smaller than the second
        - `EQUAL` if both lists are equal in contents and length
        - `LARGER` if the first list is larger than the second

        The return value represents the relation of first
        list to the second one. `Smaller` means either

        - delta values of first list are smaller -or-
        - deltas are equal but first list is shorter.

        Larger is the opposite case.

        Arguments:
            delta_list1: first monomial list to compare.
            delta_list2: second monomial list to compare.

        Returns:
            Result of comparison.
        """
        # element wise comparison up to length of shorter list
        list_diff = [(a == b) for a, b in zip(delta_list1, delta_list2)]

        # if some difference exists
        if False in list_diff:

            # index of first difference
            idx = list_diff.index(False)

            (i, j), (m, n) = delta_list1[idx], delta_list2[idx]

            if (j < n) or (j == n and i < m):
                return Comparison.SMALLER
            else:
                return Comparison.LARGER

        # If the list coincide on their initial segment up to "max",
        # determine the lengths of the two lists
        first_len, second_len = len(delta_list1), len(delta_list2)

        # determine how first list relates to second list
        if first_len > second_len:
            return Comparison.LARGER
        if first_len < second_len:
            return Comparison.SMALLER
        else:
            return Comparison.EQUAL

    @staticmethod
    def sort_monomials(monomials: list) -> list:
        """Given a list of monomials this method will return them in order.

        The sort is performed by first dividing the list of monomials into
        halves recursively until each half contains at most one monomial.
        Then the sort will begin to combine (or zip) the halves into a sorted
        list.

        The sort performs comparison of deltas, and orders the monomials based
        on the delta values. If two monomials have the same deltas, we compute
        new scalar value, and if it is not 0, we keep the result monomial.
        Note that if we get 2 monomials with same deltas, and only at most 1
        is kept, with possibly updated scalar. This means sort can return a
        result that is shorter than the input argument.

        The original list argument is not mutated by this sort operation, i.e.
        this is not sort in place.

        Arguments:
            monomials: list of monomials to sort.

        Returns:
            list of sorted monomials.
        """
        list_len = len(monomials)

        # base case
        if list_len < 2:
            if monomials and monomials[0] == ZERO_MWP:
                return []
            return monomials

        # split list into two halves equally
        mid_point = list_len // 2
        left = Polynomial.sort_monomials(monomials[mid_point:])
        right = Polynomial.sort_monomials(monomials[:mid_point])

        # construct sorted list by iteratively combining
        # monomials from left and right halves
        new_list = []
        while left and right:
            lhead, *ltail = left
            rhead, *rtail = right
            comparison = Polynomial.compare(lhead.deltas, rhead.deltas)

            # head of left list is smaller -> add it to list
            if comparison == Comparison.SMALLER:
                new_list.append(lhead)
                left = ltail

            # head of right list is smaller -> add it to list
            if comparison == Comparison.LARGER:
                new_list.append(rhead)
                right = rtail

            # list heads are equal i.e. same deltas.
            if comparison == Comparison.EQUAL:
                monomial = lhead
                # append to list as long as scalar
                # product is not 0
                monomial.scalar = sum_mwp(
                    lhead.scalar, rhead.scalar)
                if monomial.scalar != ZERO_MWP:
                    new_list.append(monomial)
                left = ltail
                right = rtail

        # either left or right is empty so order
        # doesn't matter; just append whatever
        # remains of left or right tail
        return new_list + right + left

    def remove_zeros(self) -> Polynomial:
        """Removes all encountered 0s from a polynomial.

        Before returning, if the list is empty, the result produces a
            0-monomial.

        Returns:
            polynomial with list of monomials for which zeros are
            removed, unless 0 is the only monomial.
        """
        filtered_monomials = list(filter(
            lambda mono: mono.scalar != ZERO_MWP, self.list))

        if len(filtered_monomials) == 0:
            self.list = [Monomial(ZERO_MWP)]
        else:
            self.list = filtered_monomials

        return self

    @staticmethod
    def from_scalars(index: int, *scalars: str) -> Polynomial:
        """Build a polynomial of multiple monomials with deltas.

        Example:
            For arguments `index=5` and `scalars= m, w, p`,
            the method returns a Polynomial equal to:

            ```Python
            m1 = Monomial('m', (0, 5))
            m2 = Monomial('w', (1, 5))
            m3 = Monomial('p', (2, 5))

            Polynomial(m1, m2, m3)
            ```

        Arguments:
            index: Delta index.
            scalars: Scalar values.

        Returns:
             Generated polynomial.
         """

        monomials = [Monomial(scalar, (number, index))
                     for number, scalar in enumerate(scalars)]
        return Polynomial(*monomials)



# -----------------------------------------------------------------------------
# Copyright (c) 2020-2024 C. Aubert, T. Rubiano, N. Rusch and T. Seiller.
#
# This file is part of pymwp.
#
# pymwp is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pymwp is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pymwp. If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

import setuptools
import os

__title__ = "pymwp"
__author__ = "Clément Aubert, Thomas Rubiano, Neea Rusch, Thomas Seiller"
__desc__ = "Implementation of MWP analysis on C code in Python."

with open('README.md', 'r', encoding='utf-8') as fh:
    readme = fh.read()
    start_key, end_key = '<!--include-start-->', '<!--include-end-->'
    start = readme.index(start_key) + len(start_key)
    end = readme.index(end_key)
    long_description = readme[start:end]

with open(os.path.join(os.path.abspath(os.path.dirname(__file__)),
                       "pymwp", "version.py")) as fp:
    exec(fp.read())

setuptools.setup(
    name=__title__,
    version=__version__,  # noqa: F821
    author=__author__,
    author_email='nrusch@augusta.edu',
    packages=['pymwp'],
    entry_points={'console_scripts': ['pymwp = pymwp.__main__:main'], },
    license='GPLv3+',
    description=__desc__,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/statycc/pymwp',
    project_urls={
        'Bug Tracker': 'https://github.com/statycc/pymwp/issues',
        'Documentation': 'https://statycc.github.io/pymwp/',
        'Source Code': 'https://github.com/statycc/pymwp',
        'Archive': 'https://doi.org/10.5281/zenodo.7879822'
    },
    license_files=('LICENSE',),
    package_data={"": ["LICENSE"], },
    include_package_data=True,
    classifiers=[
        'Natural Language :: English',
        'Intended Audience :: Science/Research',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Operating System :: OS Independent',
        'Environment :: Console',
        'Typing :: Typed',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
        ('License :: OSI Approved :: ' +
         'GNU General Public License v3 or later (GPLv3+)')
    ],
    python_requires=">=3.7",
    install_requires=[
        'pycparser',
        'pycparser-fake-libc'
    ]
)



#!/usr/bin/env python3

# -----------------------------------------------------------------------------
# Copyright (c) 2020-2024 C. Aubert, T. Rubiano, N. Rusch and T. Seiller.
#
# This file is part of pymwp.
#
# pymwp is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pymwp is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pymwp. If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

"""
Provides command line interface to executing pymwp analysis with arguments.

This method enables executing MWP analysis through command line.

If user has installed pymwp through pip, this will be the entry point
of that command when calling:  `pymwp c/file/path --args`

The command behavior is to run the analysis on the specified file,
applying the optional flags.

The available arguments are specified below in `_parse_args` -method.

This method also initializes the program logger that displays debugging
information on the screen. The logger default to log level DEBUG. Analysis
output can be muted by specifying command line argument `--silent`.
"""

import argparse
import logging
import sys
from argparse import RawTextHelpFormatter
from typing import List, Optional, Type, Union

from . import __version__, __title__ as pymwp
from . import __notice__, __warranty__, __conditions__
from . import Parser, Result, Analysis, LoopAnalysis
from .file_io import default_file_out, loc, save_result


def main():
    """Implementation of MWP analysis on C code in Python."""
    parser = argparse.ArgumentParser(
        prog=pymwp, description=main.__doc__ + '\n\n' + __notice__,
        formatter_class=RawTextHelpFormatter)
    args = __parse_args(parser)

    if args.license:
        import shutil
        shell_size = shutil.get_terminal_size((50, 20))
        text = __conditions__ if args.license == 'C' else __warranty__
        print(Result.pretty_print(text, shell_size.columns - 5))
        sys.exit(0)

    if not args.input_file:
        parser.print_help()
        sys.exit(1)

    # setup logger
    level = 0 if args.silent else 30 if args.info else 40
    __setup_logger(logging.FATAL - level, args.logfile, args.no_time)

    # get parser args then get AST
    parser_kwargs = {'use_cpp': not args.no_cpp}
    if not args.no_cpp:  # only when use_cpp is True
        parser_kwargs['cpp_path'] = args.cpp_path
        parser_kwargs['cpp_args'] = args.cpp_args
    c_headers = args.headers.split(',') if args.headers else None
    ast = Parser.parse(args.input_file, c_headers, **(parser_kwargs or {}))

    # setup arguments
    result = Result()
    result.program.program_path = args.input_file
    result.program.n_lines = loc(args.input_file)

    analyzer: Type[Union[Analysis, LoopAnalysis]] = \
        LoopAnalysis if args.mode == 'L' else Analysis
    result = analyzer.run(ast, result, fin=args.fin, strict=args.strict)

    if not args.no_save:
        file_out = args.out or default_file_out(args.input_file)
        save_result(file_out, result)


def __parse_args(
        parser: argparse.ArgumentParser, args: Optional[List] = None
) -> argparse.Namespace:
    """Setup available program arguments."""
    parser.add_argument(
        'input_file',
        help="C source code file to analyze",
        nargs="?"
    )
    parser.add_argument(
        '-v', "--version",
        action="version",
        version="%(prog)s " + __version__,
    )
    parser.add_argument(
        "--license",
        action='store',
        dest='license',
        type=str.upper,
        metavar="opt",
        choices=['W', 'C'],
        help="show warranty (W) or conditions (C) and exit"
    )

    analysis = parser.add_argument_group('Analysis options')
    analysis.add_argument(
        "--mode", '-m',
        action='store',
        dest='mode',
        default='F',
        type=str.upper,
        metavar="opt",
        choices=['F', 'L'],
        help="analyze functions (F) or loops (L) [default: F]"
    )
    analysis.add_argument(
        '--out', '-o',
        action="store",
        dest="out",
        metavar="FILE",
        help="file where to store analysis result",
    )
    analysis.add_argument(
        "--no_save",
        action='store_true',
        help="do not write analysis result to a file"
    )
    analysis.add_argument(
        "--fin",
        action='store_true',
        help="ensure analysis always runs to completion"
    )
    analysis.add_argument(
        "--strict",
        action='store_true',
        help="enforce input file must be syntax-compliant"
    )

    compiler = parser.add_argument_group('C compiler options')
    compiler.add_argument(
        '--cpp_path',
        action='store',
        default='gcc',
        metavar="PATH",
        help='pre-processor [default: gcc]',
    )
    compiler.add_argument(
        '--cpp_args',
        action='store',
        default='-E',
        metavar="ARGS",
        help='pre-processor arguments [default: -E]',
    )
    compiler.add_argument(
        "--headers",
        action="store",
        metavar="DIR",
        help="headers dir paths, separate by comma",
    )
    compiler.add_argument(
        "--no_cpp",
        action='store_true',
        help="disable pre-processor"
    )

    log_group = parser.add_argument_group('Terminal log options')
    log_group.add_argument(
        "--logfile",
        action="store",
        metavar="FILE",
        help="write console output to a file",
    )
    log_group.add_argument(
        "--no_time",
        action='store_true',
        help="omits timestamps from log output"
    )
    log_group.add_argument(
        '--info',
        action='store_true',
        help="set logging level to info"
    )
    log_group.add_argument(
        "--silent",
        action='store_true',
        help="disable all terminal output"
    )

    return parser.parse_args(args)


def __setup_logger(
        level: int = logging.ERROR, log_filename: Optional[str] = None,
        hide_time: bool = False
) -> None:
    """Create a configured instance of logger.

    Arguments:
        level: Describe the severity level of the logs to handle.
            see: https://docs.python.org/3/library/logging.html#levels
        log_filename: Write logging info to a file
    """
    fmt = "%(levelname)s (%(module)s): %(message)s" if hide_time else \
        "[%(asctime)s] %(levelname)s (%(module)s): %(message)s"
    date_fmt = "%H:%M:%S"
    formatter = logging.Formatter(fmt, datefmt=date_fmt)

    logger = logging.getLogger(pymwp)
    logger.setLevel(level)

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)

    if log_filename is not None:
        file_handler = logging.FileHandler(log_filename)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)


if __name__ == '__main__':
    main()



# -----------------------------------------------------------------------------
# Copyright (c) 2020-2024 C. Aubert, T. Rubiano, N. Rusch and T. Seiller.
#
# This file is part of pymwp.
#
# pymwp is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pymwp is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pymwp. If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

from __future__ import annotations
import os
import re
import tempfile
from abc import ABC, abstractmethod
from logging import getLogger
# noinspection PyPackageRequirements,PyProtectedMember
from typing import Any, List, Type, Union

# noinspection PyProtectedMember
from pycparser import c_ast, parse_file, c_generator
from pycparser_fake_libc import directory as fake_libc_dir

logger = getLogger(__name__)

"""Interface the pycparser dependency.

Make all explicit references to pycparser within this module.
If some functionality is missing, extend this interface as necessary,
then test the used functionality in tests/test_parser.py.
Do not import pycparser directly elsewhere outside this module+test.

The motivation for using this interface is to make it easier to
test and replace this dependency with newer versions and ensuring it
such upgrades do not break pymwp.
"""


# noinspection PyPep8Naming
class Nodes(ABC):  # pragma: no cover
    @property
    @abstractmethod
    def ArrayDecl(self):
        pass

    @property
    @abstractmethod
    def ArrayRef(self):
        pass

    @property
    @abstractmethod
    def Assignment(self):
        pass

    @property
    @abstractmethod
    def BinaryOp(self):
        pass

    @property
    @abstractmethod
    def Break(self):
        pass

    @property
    @abstractmethod
    def Case(self):
        pass

    @property
    @abstractmethod
    def Cast(self):
        pass

    @property
    @abstractmethod
    def Compound(self):
        pass

    @property
    @abstractmethod
    def Constant(self):
        pass

    @property
    @abstractmethod
    def Continue(self):
        pass

    @property
    @abstractmethod
    def Decl(self):
        pass

    @property
    @abstractmethod
    def DeclList(self):
        pass

    @property
    @abstractmethod
    def Default(self):
        pass

    @property
    @abstractmethod
    def DoWhile(self):
        pass

    @property
    @abstractmethod
    def EmptyStatement(self):
        pass

    @property
    @abstractmethod
    def ExprList(self):
        pass

    @property
    @abstractmethod
    def For(self):
        pass

    @property
    @abstractmethod
    def FuncCall(self):
        pass

    @property
    @abstractmethod
    def FuncDef(self):
        pass

    @property
    @abstractmethod
    def ID(self):
        pass

    @property
    @abstractmethod
    def If(self):
        pass

    @property
    @abstractmethod
    def ParamList(self):
        pass

    @property
    @abstractmethod
    def Return(self):
        pass

    @property
    @abstractmethod
    def Switch(self):
        pass

    @property
    @abstractmethod
    def TernaryOp(self):
        pass

    @property
    @abstractmethod
    def TypeDecl(self):
        pass

    @property
    @abstractmethod
    def UnaryOp(self):
        pass

    @property
    @abstractmethod
    def While(self):
        pass


# noinspection PyPep8Naming
class NodeHandler(ABC):  # pragma: no cover

    def ArrayDecl(self, node: Nodes.ArrayDecl, *args, **kwargs):
        pass

    def ArrayRef(self, node: Nodes.ArrayRef, *args, **kwargs):
        pass

    def Assignment(self, node: Nodes.Assignment, *args, **kwargs):
        pass

    def BinaryOp(self, node: Nodes.BinaryOp, *args, **kwargs):
        pass

    def Break(self, node: Nodes.Break, *args, **kwargs):
        pass

    def Case(self, node: Nodes.Case, *args, **kwargs):
        pass

    def Cast(self, node: Nodes.Cast, *args, **kwargs):
        pass

    def Compound(self, node: Nodes.Compound, *args, **kwargs):
        pass

    def Constant(self, node: Nodes.Constant, *args, **kwargs):
        pass

    def Continue(self, node: Nodes.Continue, *args, **kwargs):
        pass

    def Decl(self, node: Nodes.Decl, *args, **kwargs):
        pass

    def DeclList(self, node: Nodes.DeclList, *args, **kwargs):
        pass

    def Default(self, node: Nodes.Default, *args, **kwargs):
        pass

    def DoWhile(self, node: Nodes.DoWhile, *args, **kwargs):
        pass

    def EmptyStatement(self, node: Nodes.EmptyStatement, *args, **kwargs):
        pass

    def ExprList(self, node: Nodes.ExprList, *args, **kwargs):
        pass

    def For(self, node: Nodes.For, *args, **kwargs):
        pass

    def FuncCall(self, node: Nodes.FuncCall, *args, **kwargs):
        pass

    def FuncDef(self, node: Nodes.FuncDef, *args, **kwargs):
        pass

    def ID(self, node: Nodes.ID, *args, **kwargs):
        pass

    def If(self, node: Nodes.If, *args, **kwargs):
        pass

    def ParamList(self, node: Nodes.ParamList, *args, **kwargs):
        pass

    def Return(self, node: Nodes.Return, *args, **kwargs):
        pass

    def Switch(self, node: Nodes.Switch, *args, **kwargs):
        pass

    def TernaryOp(self, node: Nodes.TernaryOp, *args, **kwargs):
        pass

    def TypeDecl(self, node: Nodes.TypeDecl, *args, **kwargs):
        pass

    def UnaryOp(self, node: Nodes.UnaryOp, *args, **kwargs):
        pass

    def While(self, node: Nodes.While, *args, **kwargs):
        pass


# noinspection PyPep8Naming
class ParserInterface(Nodes):  # pragma: no cover
    """Interface for C code parser."""

    @staticmethod
    @abstractmethod
    def parse(full_file_name: str, **kwargs) -> dict:
        """Extract text from the currently loaded file."""
        pass

    @abstractmethod
    def is_func(self, node: Any) -> bool:
        """True if node is a function implementation."""
        pass

    @abstractmethod
    def is_loop(self, node: Any) -> bool:
        """True is node is a loop statement."""
        pass

    @abstractmethod
    def to_c(self, node: Any) -> str:
        """Translate node back to C code."""
        pass

    @property
    @abstractmethod
    def Node(self) -> Type:
        """Base type for all AST nodes."""
        pass

    # noinspection PyPep8Naming
    @property
    @abstractmethod
    def LoopT(self) -> Type:
        """Loop type."""
        pass


class PyCParser(ParserInterface):
    """Implementation of the parser interface using pycparser."""

    @staticmethod
    def parse(file_name: str, headers: List[str] = None, **kwargs):

        # build parser cpp arguments
        # always append -E and fake_libc args when C preprocessor
        # is enabled
        if kwargs.get('use_cpp', False):
            args = kwargs.get('cpp_args', None)
            args = args.split(' ') if args else []
            if '-E' not in args:
                args.append(r'-E')
            args.append(r'-I' + fake_libc_dir)
            if headers:
                for h in headers:
                    args.append(r'-I' + h)
            kwargs['cpp_args'] = args
        logger.debug(f'parser arguments: {kwargs}')

        # if running windows there is no automated fix for you, boo hoo
        # add headers manually

        if os.name == 'nt':
            warning = 'automatic pre-parser processing was not ' \
                      'applied on Windows: manual fix may be needed'
            logger.warning(warning)
            # on Windows parse the original file as-is
            return parse_file(file_name, **kwargs)

        # read the original C program source
        with open(file_name, "r+") as cfile:
            c_prog = cfile.read()

        # apply preprocessing steps
        preprocessed = PyCParser.add_attr_x(c_prog)
        # convert to byte string
        prog_bytes = str.encode(preprocessed)

        # create temporary file copy where we apply custom preprocessing
        fp = tempfile.NamedTemporaryFile(suffix=".c")
        fp.write(prog_bytes)
        fp.seek(0)
        # pass the temporary file pointer to pycparser
        ast = parse_file(fp.name, **kwargs)
        fp.close()
        return ast

    def is_func(self, node: ParserInterface.Node) -> bool:
        """True if node is a (non-empty) function implementation."""
        return (isinstance(node, self.FuncDef) and
                hasattr(node, 'body') and
                hasattr(node.body, 'block_items'))

    def is_loop(self, node: ParserInterface.Node) -> bool:
        """True is node is a (non-empty) loop statement."""
        return (isinstance(node, (self.While, self.For, self.DoWhile))
                and hasattr(node, 'stmt') and node.stmt
                and not isinstance(node.stmt, self.EmptyStatement))

    def to_c(self, node: Any, compact: bool = False) -> str:
        """Translate node back to C code."""
        generator = c_generator.CGenerator()
        comm = generator.visit(node)
        if compact:
            comm = re.sub(r"[\n\t\s]+", " ", comm).strip()
        return comm.strip()

    @property
    def Node(self) -> Type:
        """Base type for all AST nodes."""
        return Type[c_ast.Node]

    # noinspection PyPep8Naming
    @property
    def LoopT(self) -> Type:
        """Loop type."""
        return Type[Union[Parser.While, Parser.DoWhile, Parser.For]]

    @staticmethod
    def add_attr_x(text: str) -> str:
        """Conditionally add `#define __attribute__(x)` to C file
        for pycparser.

        See: <https://github.com/eliben/pycparser/wiki/FAQ#what-do-i-do
        -about-__attribute__/>

        Arguments:
            text: C program file content as a string

        Returns:
            contents of C file, with `#define __attribute__(x)` included.
        """
        attr_x = '#define __attribute__(x)'
        lines = text.split('\n')
        not_found = not any([line.startswith(attr_x) for line in lines])
        if not_found:
            lines.insert(0, attr_x)
            text = '\n'.join(lines)
            logger.debug(f'inserted {attr_x}')
        return text

    @property
    def ArrayDecl(self):
        return c_ast.ArrayDecl

    @property
    def ArrayRef(self):
        return c_ast.ArrayRef

    @property
    def AST(self):
        return c_ast

    @property
    def Assignment(self):
        return c_ast.Assignment

    @property
    def BinaryOp(self):
        return c_ast.BinaryOp

    @property
    def Break(self):
        return c_ast.Break

    @property
    def Case(self):
        return c_ast.Case

    @property
    def Cast(self):
        return c_ast.Cast

    @property
    def Compound(self):
        return c_ast.Compound

    @property
    def Constant(self):
        return c_ast.Constant

    @property
    def Continue(self):
        return c_ast.Continue

    @property
    def Decl(self):
        return c_ast.Decl

    @property
    def DeclList(self):
        return c_ast.DeclList

    @property
    def Default(self):
        return c_ast.Default

    @property
    def DoWhile(self):
        return c_ast.DoWhile

    @property
    def ExprList(self):
        return c_ast.ExprList

    @property
    def EmptyStatement(self):
        return c_ast.EmptyStatement

    @property
    def For(self):
        return c_ast.For

    @property
    def FuncCall(self):
        return c_ast.FuncCall

    @property
    def FuncDef(self):
        return c_ast.FuncDef

    @property
    def ID(self):
        return c_ast.ID

    @property
    def If(self):
        return c_ast.If

    @property
    def ParamList(self):
        return c_ast.ParamList

    @property
    def Return(self):
        return c_ast.Return

    @property
    def Switch(self):
        return c_ast.Switch

    @property
    def TernaryOp(self):
        return c_ast.TernaryOp

    @property
    def TypeDecl(self):
        return c_ast.TypeDecl

    @property
    def UnaryOp(self):
        return c_ast.UnaryOp

    @property
    def While(self):
        return c_ast.While


# Parser is an instance of the preferred parser implementation
Parser = PyCParser()



# -----------------------------------------------------------------------------
# Copyright (c) 2020-2024 C. Aubert, T. Rubiano, N. Rusch and T. Seiller.
#
# This file is part of pymwp.
#
# pymwp is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pymwp is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pymwp. If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# flake8: noqa: W605

from __future__ import annotations

from typing import List, Optional, Tuple, Union

from . import DELTA
from .constants import SetInclusion
from .semiring import ZERO_MWP, UNIT_MWP, prod_mwp, sum_mwp


class Monomial:
    """
    A monomial is a pair made of:

    1. `scalar` - a value in the semi-ring
    2. a sorted list of `deltas`, where an index occurs at most once.

    Deltas are coded as pairs $(i,j)$ with:

    - $i$ the value and
    - $j$ the index in the domain (infinite product)

    We will have that $(i,j)$ will be equal to the unit of the semi-ring
    iff the $j^{th}$ input is equal to $i$ (so, the $j^{th}$ choice is $i$).

    We will make the assumption that the deltas of delta is sorted
    and no two deltas can have the same index.

    Attributes:
        scalar (str): Monomial scalar.
        deltas (List[DELTA]): List of deltas.
    """

    def __init__(self, scalar: str = UNIT_MWP,
                 deltas: Optional[Union[List[DELTA], DELTA]] = None,
                 *args: Optional[DELTA]):
        """Create a monomial.

        Example:
            Create a monomial.

            ```python
            mono = Monomial()
            ```

            Create monomial with scalar $m$ explicitly.

            ```python
            mono = Monomial('m')
            ```

            Create monomial with scalar $w$ and two deltas.

            ```python
            mono = Monomial('w', (0, 0), (1, 1))
            ```

        Arguments:
            scalar: Monomial scalar.
            deltas: A delta or a list of deltas.
            *args: Arbitrary number of subsequent deltas.
        """
        self.deltas = []
        self.scalar = scalar

        if deltas is not None:
            if isinstance(deltas, list):
                Monomial.insert_deltas(self, deltas)
            if isinstance(deltas, tuple):
                delta_list = [deltas] + list(args if args else [])
                Monomial.insert_deltas(self, delta_list)

    def __str__(self) -> str:
        deltas = [".delta({0},{1})".format(*delta) for delta in self.deltas]
        return self.scalar + ''.join(deltas)

    def __mul__(self, other) -> Monomial:
        return self.prod(other)

    @staticmethod
    def format(value: Union[str, Monomial, Tuple]) -> Monomial:
        if isinstance(value, Monomial):
            return value
        if isinstance(value, str):
            return Monomial(value)
        if isinstance(value, Tuple):
            return Monomial(*value)

    def contains(self, m: Monomial) -> bool:
        """Check if all deltas of m are in deltas of self.

        Arguments:
            m: A monomial to search for intersection.

        Returns:
            False if one delta of m not in self, True otherwise.
        """
        for b in m.deltas:
            if b not in self.deltas:
                return False
        return True

    def inclusion(self, monomial: Monomial) -> SetInclusion:
        """Gives info about inclusion of self monomial with monomial.

        Arguments:
            monomial: A monomial to see inclusion.

        Returns:
            CONTAINS if self contains monomial, INCLUDED if self is included
                in monomial, and EMPTY none of them.
        """
        # self contains monomial ?
        contains: bool = self.contains(monomial)

        summ = sum_mwp(self.scalar, monomial.scalar)
        # if self contains monomial and self.scalar >= monomial.scalar
        if contains and (monomial.scalar == summ):
            return SetInclusion.CONTAINS
        else:
            # self included in monomial and self.scalar <= monomial.scalar
            included: bool = monomial.contains(self)
            if included and (self.scalar == summ):
                return SetInclusion.INCLUDED
            else:
                return SetInclusion.EMPTY

    def prod(self, monomial: Monomial) -> Monomial:
        """Prod combines two monomials where one is this monomial (self)
        and the second is an argument.

        The attributes of the resulting monomial are determined as follows:

        - output scalar is a product of the input scalars
        - two lists of deltas are merged according to rules of insert_delta

        Arguments:
            monomial: The second monomial.

        Returns:
            A new Monomial that is a product of two monomials.
        """
        # make a copy of self
        mono_product = self.copy()

        # determine the new scalar
        mono_product.scalar = prod_mwp(self.scalar, monomial.scalar)

        # if scalar is 0, monomial cannot have deltas
        if mono_product.scalar == ZERO_MWP:
            mono_product.deltas = []

        # otherwise merge the two lists of deltas
        # result already contains deltas from "self"
        # so we are adding to it the deltas from
        # the second monomial passed in as argument
        elif monomial.deltas:
            # TODO here insert only those not contained ?
            Monomial.insert_deltas(mono_product, monomial.deltas)

        return mono_product

    def choice_scalar(self, *choices: int) -> Optional[str]:
        """Determine if given sequence of choices matches monomial.

        Arguments:
            choices: Tuple of choices.

        Returns:
            Monomial's scalar if structure matches choices and None otherwise.
        """
        for (i, j) in self.deltas:
            if not i == choices[j]:
                return None
        return self.scalar

    def copy(self) -> Monomial:
        """Make a deep copy of a monomial."""
        return Monomial(self.scalar, self.deltas[:])

    def show(self) -> None:
        """Display scalar and the list of deltas."""
        print(str(self))

    def to_dict(self) -> dict:
        """Get dictionary representation of a monomial."""
        return {
            "scalar": self.scalar,
            "deltas": self.deltas
        }

    @staticmethod
    def insert_deltas(monomial: Monomial, deltas: List[DELTA]) -> None:
        """Insert new deltas into monomial list of deltas.

        Arguments:
            monomial: The monomial into whose list of deltas values will
                be inserted.
            deltas: List of deltas to insert into monomial.
        """

        # Deltas are inserted one after the other so that
        # the resulting list is ordered
        for delta in deltas:
            monomial.deltas = Monomial.insert_delta(monomial.deltas, delta)

            # change the scalar to 0 because we have a null
            # monomial after insert
            if not monomial.deltas:
                monomial.scalar = ZERO_MWP
                # we can stop inserting if this happens
                break

    @staticmethod
    def insert_delta(sorted_deltas: List[DELTA], delta: DELTA) -> List[DELTA]:
        """
        Takes as input a _sorted_ list of deltas and a delta.

        Check if two deltas have the same index:

        If they do, and if they:

        - disagree on the value expected, returns `[]` (empty list)
        - agree on the value expected, returns the original deltas

        If they don't:

        - add the new delta in the deltas "at the right position".

        Arguments:
            sorted_deltas: List of deltas where to perform insert.
            delta: The delta value to be inserted.

        Returns:
            updated list of deltas.
        """

        # insert position index
        i = 0

        # iterate existing deltas and look for
        # where to insert the new value
        while i < len(sorted_deltas):

            # if delta to insert has higher index
            # go to next one
            if sorted_deltas[i][1] < delta[1]:
                i = i + 1

            # if delta to insert matches existing index
            elif sorted_deltas[i][1] == delta[1]:

                # when delta to insert is already in the
                # list, return the original list without
                # performing insert
                if sorted_deltas[i][0] == delta[0]:
                    return sorted_deltas

                # If the delta disagrees with the choices
                # previously stored, we simply return the
                # empty list: we will never be able to
                # accommodate both the requirement of
                # the existing list and of the new delta.
                return []

            else:
                break

        # perform insert at appropriate index
        sorted_deltas.insert(i, delta)
        return sorted_deltas
